//
// Copyright (c) 2008 EmbeddedDreams.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the
// Free Software Foundation, Inc.,
// 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// $Id: selfdcc.c,v 1.2 2008/08/20 00:45:17 nsj Exp $
//

#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/sleep.h>
#include <stdlib.h>
#include <avr/io.h>
#include "tslib.h"


///////////////////////////////////////////////////////////////////////////
// Macros

#define ROUND(x)            ((int)((x) + 0.5))

// Voltage output is checked at this frequency and, if necessary, corrected.
#define TIMER0_FREQ         10000
#define TIMER0_PER_x10US    ((100000 + TIMER0_FREQ/2) / TIMER0_FREQ)

// Softstart will stop when output reaches Vout minus the value below.
#define SOFTSTART_STOP      0.0     // V
// PWM duty cycle change frequency during soft start.
#define SOFTSTARTSTEP_FREQ  1000
#define SOFTSTARTSTEPK      ((TIMER0_FREQ + SOFTSTARTSTEP_FREQ/2) / SOFTSTARTSTEP_FREQ)

#define CTRCLK              64000000UL      // or F_CPU
#define PWM_FREQ            62500
#define PWM_PRESCALER       4
#define PWM_DIVIDER         ROUND(1.0 * CTRCLK / PWM_PRESCALER / PWM_FREQ - 1)
// Limit the maximum PWM duty cycle.
#define MAX_DUTYCYCLE       0.996

#define VREF                2.7     // AVR internal Vref used for ADC.
#define VOUT                5.0     // target regulator output voltage.
// VOUT is scaled down by an external resistor divider, by the amount below.
#define FB_SCALE            0.5
// VOUT * FB_SCALE = adc * VREF / 256 => adc = VOUT * FB_SCALE * 256 / VREF
#define FB_ADC_COUNT        ROUND(VOUT * FB_SCALE * (256*2) / VREF)
// VOUT variation corresponding to 1 adc LSB.
#define VOUT_LSB_DELTA      (VREF / (256.0*2) / FB_SCALE)


#ifdef SIMULATION
#   define _delay_us(x)
#   undef sleep_mode
#   define sleep_mode()
#endif


///////////////////////////////////////////////////////////////////////////
// Types

#define true    (1 == 1)
#define false   (0 == 1)

// PORT A bit meanings
typedef enum {
    eOUT_SENSE  = _BV(PA0),
    eCheckVout  = _BV(PA4),
} e_porta_bits;

// PORT B bit meanings
typedef enum {
    ePWM       = _BV(PB1),     // Frequency Output
} e_portb_bits;


///////////////////////////////////////////////////////////////////////////
// Global variables

static byte gSoftStart = SOFTSTARTSTEPK;


///////////////////////////////////////////////////////////////////////////
// Code

static byte gTol = 1;

ISR(ADC_vect)
{
    portA_SetPin(4, true);

    word  adc = adc_Read10() / 2;
    byte  duty = OCR1A;
    byte  delta = 1;

    if (gSoftStart)
    {
        // Stop softstart when the regulated voltage reaches a certain
        // value near (below) VOUT.
        if (adc >= FB_ADC_COUNT - ROUND(SOFTSTART_STOP / VOUT_LSB_DELTA))
        {
            gSoftStart = 0;
        }
        else
        {
            gSoftStart--;
            if (gSoftStart != 1)
            {
                return;
            }
            gSoftStart = SOFTSTARTSTEPK;
        }
    }

    if (adc < FB_ADC_COUNT - gTol)
    {
        if (duty < PWM_DIVIDER - delta)
        {
            duty += delta;
        }
    }
    else if (adc > FB_ADC_COUNT + 0)
    {
        if (duty > delta)
        {
            duty -= delta;
        }
    }
    else
    {
        // perfect!! no corrections to perform
    }
    OCR1A = duty;

    portA_SetPin(4, false);
}


// Periodically check the output voltage and adjust accordingly.
TIMER0_HANDLER()

    adc_TriggerSingleConversion();

TIMER0_HANDLER_END


static byte calcPrescaler (word divide)
{
    switch (divide)
    {
        default  : return 0;
        case 1   : return 1;
        case 2   : return 2;
        case 4   : return 3;
        case 8   : return 4;
        case 16  : return 5;
        case 32  : return 6;
        case 64  : return 7;
        case 128 : return 8;
        case 256 : return 9;
        case 512 : return 10;
        case 1024: return 11;
        case 2048: return 12;
        // ...
    }
}


/**
 * Initialize the controller and the application.
 */
static void avr_init (void)
{
    // Configure output ports.
    DDRA = eCheckVout;
    DDRB = ePWM;
    led_Setup (eLedAll);

    // Program counter/PWM
    OCR1C = PWM_DIVIDER;
    OCR1A = 32;          // start almost quietly
    TCCR1B |= calcPrescaler(PWM_PRESCALER);     // Clock pre-scaler
    TCCR1A |= _BV(PWM1A);       // Enable PWM
    TCCR1A |= _BV(COM1A1);      // Use OC1A output (PB1)

    // Enable fast PWM clock
    PLLCSR |= _BV(PLLE);        // Enable the fast 64MHz PLL clock.
    //_delay_us(150);             // Wait at least 100us for PLL stabilization.
    while (!(PLLCSR & _BV(PLOCK)));  // Wait for the PLL to lock. Must be done before PCKE.
    PLLCSR |= _BV(PCKE);        // Enable asynch mode (select 64MHz clock).*/

    timer0_SetupPeriodicHz (TIMER0_FREQ);

    adc_Setup (10000);
    adc_CfgInputPin (eADC0, eSingleEnd);
    adc_SetVRef (e2V56);   // 2.7V nominal
    adc_Enable10bitMode();

    // Set sleep mode to "Idle"
    set_sleep_mode(0);
    but_Setup(eButAll);
}


/**
 * Application entry point.
 */
int __attribute__((noreturn)) main (void)
{
    avr_init();

    // Enable interrupts!!
    sei();

    for (;;)
    {
        static word c = 0;
        word  i;

        c++;
        if (c == TIMER0_FREQ/2)
        {
            led_Toggle(eLedGreen);      // toggle an "I'm alive" LED
            c = 0;
        }

        // block in low-power mode until after the next timer interrupt occurs
        Sleep();
        // Adjust voltage tolerance with 2 buttons
        if (but_IsPressed(eBut1))
        {
            for (i = 0; i < TIMER0_FREQ/20; i--) Sleep();
            while (but_IsPressed(eBut1));
            //OCR1A ++;
            gTol++;
        }
        if (but_IsPressed(eBut2))
        {
            for (i = 0; i < TIMER0_FREQ/20; i--) Sleep();
            while (but_IsPressed(eBut2));
            //OCR1A --;
            gTol--;
        }
    }
}

